from __future__ import annotations

import sys
import typing
from dataclasses import dataclass
from inspect import Parameter, Signature, getmembers, isclass, ismethod
from itertools import chain
from typing import Any, AnyStr, Collection, ForwardRef, TypeVar, Union

from typing_extensions import Annotated, NotRequired, Required, Self, get_args, get_origin, get_type_hints

from litestar import connection, datastructures, types
from litestar.datastructures import ImmutableState
from litestar.enums import RequestEncodingType
from litestar.exceptions import ImproperlyConfiguredException
from litestar.params import BodyKwarg, DependencyKwarg, ParameterKwarg
from litestar.types import AnyCallable, Empty
from litestar.types.builtin_types import UNION_TYPES, NoneType
from litestar.utils.typing import get_safe_generic_origin, unwrap_annotation

_GLOBAL_NAMES = {
    namespace: export
    for namespace, export in chain(
        tuple(getmembers(types)), tuple(getmembers(connection)), tuple(getmembers(datastructures))
    )
    if namespace[0].isupper()
    and namespace in chain(types.__all__, connection.__all__, datastructures.__all__)  # pyright: ignore
}
"""A mapping of names used for handler signature forward-ref resolution.

This allows users to include these names within an `if TYPE_CHECKING:` block in their handler module.
"""

__all__ = (
    "get_fn_type_hints",
    "ParsedType",
    "ParsedParameter",
    "ParsedSignature",
    "infer_request_encoding_from_parameter",
)


def get_fn_type_hints(fn: Any, namespace: dict[str, Any] | None = None) -> dict[str, Any]:
    """Resolve type hints for ``fn``.

    Args:
        fn: Callable that is being inspected
        namespace: Extra names for resolution of forward references.

    Returns:
        Mapping of names to types.
    """
    fn_to_inspect: Any = fn

    module_name = fn_to_inspect.__module__

    if isclass(fn_to_inspect):
        fn_to_inspect = fn_to_inspect.__init__

    # detect objects that are not functions and that have a `__call__` method
    if callable(fn_to_inspect) and ismethod(fn_to_inspect.__call__):
        fn_to_inspect = fn_to_inspect.__call__

    # inspect the underlying function for methods
    if hasattr(fn_to_inspect, "__func__"):
        fn_to_inspect = fn_to_inspect.__func__

    # Order important. If a litestar name has been overridden in the function module, we want
    # to use that instead of the litestar one.
    namespace = {
        **_GLOBAL_NAMES,
        **vars(typing),
        **vars(sys.modules[module_name]),
        **(namespace or {}),
    }
    return get_type_hints(fn_to_inspect, globalns=namespace, include_extras=True)


@dataclass(frozen=True, init=False)
class ParsedType:
    """Represents a type annotation."""

    __slots__ = (
        "raw",
        "annotation",
        "origin",
        "args",
        "metadata",
        "is_annotated",
        "is_required",
        "is_not_required",
        "safe_generic_origin",
        "inner_types",
    )

    raw: Any
    """The annotation exactly as received."""
    annotation: Any
    """The annotation with any "wrapper" types removed, e.g. Annotated."""
    origin: Any
    """The result of calling ``get_origin(annotation)`` after unwrapping Annotated, e.g. list."""
    args: tuple[Any, ...]
    """The result of calling ``get_args(annotation)`` after unwrapping Annotated, e.g. (int,)."""
    metadata: tuple[Any, ...]
    """Any metadata associated with the annotation via ``Annotated``."""
    is_annotated: bool
    """Whether the annotation included ``Annotated`` or not."""
    is_required: bool
    """Whether the annotation included ``Required`` or not."""
    is_not_required: bool
    """Whether the annotation included ``NotRequired`` or not."""
    safe_generic_origin: Any
    """An equivalent type to ``origin`` that can be safely used as a generic type across all supported Python versions.

    This is to serve safely rebuilding a generic outer type with different args at runtime.
    """
    inner_types: tuple[ParsedType, ...]
    """The type's generic args parsed as ``ParsedType``, if applicable."""

    def __init__(self, annotation: Any) -> None:
        """Initialize ParsedType.

        Args:
            annotation: The type annotation. This should be extracted from the return of
                ``get_type_hints(..., include_extras=True)`` so that forward references are resolved and recursive
                ``Annotated`` types are flattened.

        Returns:
            ParsedType
        """
        unwrapped, metadata, wrappers = unwrap_annotation(annotation)
        origin = get_origin(unwrapped)
        args = get_args(unwrapped)
        object.__setattr__(self, "raw", annotation)
        object.__setattr__(self, "annotation", unwrapped)
        object.__setattr__(self, "origin", origin)
        object.__setattr__(self, "args", args)
        object.__setattr__(self, "metadata", metadata)
        object.__setattr__(self, "is_annotated", Annotated in wrappers)
        object.__setattr__(self, "is_required", Required in wrappers)
        object.__setattr__(self, "is_not_required", NotRequired in wrappers)
        object.__setattr__(self, "safe_generic_origin", get_safe_generic_origin(origin))
        object.__setattr__(self, "inner_types", tuple(ParsedType(arg) for arg in args))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ParsedType):
            return False

        if self.origin:
            return bool(self.origin == other.origin and self.inner_types == other.inner_types)

        return bool(self.annotation == other.annotation)

    @property
    def is_forward_ref(self) -> bool:
        """Whether the annotation is a forward reference or not."""
        return isinstance(self.annotation, (str, ForwardRef))

    @property
    def is_type_var(self) -> bool:
        """Whether the annotation is a TypeVar or not."""
        return isinstance(self.annotation, TypeVar)

    @property
    def is_union(self) -> bool:
        """Whether the annotation is a union type or not."""
        return self.origin in UNION_TYPES

    @property
    def is_optional(self) -> bool:
        """Whether the annotation is Optional or not."""
        return bool(self.is_union and NoneType in self.args)

    @property
    def is_collection(self) -> bool:
        """Whether the annotation is a collection type or not."""
        return bool(self.origin and self.origin is not Union and issubclass(self.origin, Collection))

    def is_subclass_of(self, cl: type[Any] | tuple[type[Any], ...]) -> bool:
        """Whether the annotation is a subclass of the given type.

        Where ``self.annotation`` is a union type, this method will always return ``False``. While this is not
        strictly correct, we intend on revisiting this once a concrete use-case is to hand.

        Args:
            cl: The type to check, or tuple of types. Passed as 2nd argument to ``issubclass()``.

        Returns:
            Whether the annotation is a subtype of the given type(s).
        """
        if self.origin:
            return self.origin not in UNION_TYPES and issubclass(self.origin, cl)
        if self.annotation is AnyStr:
            return issubclass(str, cl) or issubclass(bytes, cl)
        return self.annotation is not Any and not self.is_type_var and issubclass(self.annotation, cl)

    def has_inner_subclass_of(self, cl: type[Any] | tuple[type[Any], ...]) -> bool:
        """Whether any generic args are a subclass of the given type.

        Args:
            cl: The type to check, or tuple of types. Passed as 2nd argument to ``issubclass()``.

        Returns:
            Whether any of the type's generic args are a subclass of the given type.
        """
        return any(t.is_subclass_of(cl) for t in self.inner_types)


@dataclass(frozen=True)
class ParsedParameter:
    """Represents the parameters of a callable."""

    __slots__ = (
        "name",
        "default",
        "parsed_type",
    )

    name: str
    """The name of the parameter."""
    default: Any | Empty
    """The default value of the parameter."""
    parsed_type: ParsedType
    """The annotation of the parameter."""

    @property
    def kwarg_container(self) -> ParameterKwarg | BodyKwarg | DependencyKwarg | None:
        """A kwarg container, if any"""
        for value in (*self.metadata, self.default):
            if isinstance(value, (ParameterKwarg, BodyKwarg, DependencyKwarg)):
                return value
        return None

    @property
    def metadata(self) -> tuple[Any, ...]:
        """The metadata of the parameter's annotation."""
        return self.parsed_type.metadata

    @property
    def annotation(self) -> Any:
        """The annotation of the parameter."""
        return self.parsed_type.annotation

    @property
    def has_default(self) -> bool:
        """Whether the parameter has a default value or not."""
        return self.default is not Empty

    @classmethod
    def from_parameter(cls, parameter: Parameter, fn_type_hints: dict[str, Any]) -> ParsedParameter:
        """Initialize ParsedSignatureParameter.

        Args:
            parameter: inspect.Parameter
            fn_type_hints: mapping of names to types. Should be result of ``get_type_hints()``, preferably via the
            :attr:``get_fn_type_hints() <.utils.signature_parsing.get_fn_type_hints>` helper.

        Returns:
            ParsedSignatureParameter.
        """
        try:
            annotation = fn_type_hints[parameter.name]
        except KeyError as err:
            raise ImproperlyConfiguredException(
                f"'{parameter.name}' does not have a type annotation. If it should receive any value, use 'Any'."
            ) from err

        if parameter.name == "state" and not issubclass(annotation, ImmutableState):
            raise ImproperlyConfiguredException(
                f"The type annotation `{annotation}` is an invalid type for the 'state' reserved kwarg. "
                "It must be typed to a subclass of `litestar.datastructures.ImmutableState` or "
                "`litestar.datastructures.State`."
            )

        return ParsedParameter(
            name=parameter.name,
            default=Empty if parameter.default is Signature.empty else parameter.default,
            parsed_type=ParsedType(annotation),
        )


@dataclass(frozen=True)
class ParsedSignature:
    """Parsed signature.

    This object is the primary source of handler/dependency signature information.

    The only post-processing that occurs is the conversion of any forward referenced type annotations.
    """

    __slots__ = ("parameters", "return_type", "original_signature")

    parameters: dict[str, ParsedParameter]
    """A mapping of parameter names to ParsedSignatureParameter instances."""
    return_type: ParsedType
    """The return annotation of the callable."""
    original_signature: Signature
    """The raw signature as returned by :func:`inspect.signature`"""

    @classmethod
    def from_fn(cls, fn: AnyCallable, signature_namespace: dict[str, Any]) -> Self:
        """Parse a function signature.

        Args:
            fn: Any callable.
            signature_namespace: mapping of names to types for forward reference resolution

        Returns:
            ParsedSignature
        """
        signature = Signature.from_callable(fn)
        fn_type_hints = get_fn_type_hints(fn, namespace=signature_namespace)

        parameters = (
            ParsedParameter.from_parameter(parameter=parameter, fn_type_hints=fn_type_hints)
            for name, parameter in signature.parameters.items()
            if name not in ("self", "cls")
        )
        return cls(
            parameters={p.name: p for p in parameters},
            return_type=ParsedType(fn_type_hints.get("return", Empty)),
            original_signature=signature,
        )

    @classmethod
    def from_signature(cls, signature: Signature, signature_namespace: dict[str, Any]) -> Self:
        """Parse an :class:`inspect.Signature` instance.

        Python's `get_type_hints()` function does not support parsing signatures directly, so we need to create a dummy
        function to pass to it. Maybe there's a better way to do this, but this does work.

        Args:
            signature: An :class:`inspect.Signature` instance.
            signature_namespace: mapping of names to types for forward reference resolution

        Returns:
            ParsedSignature
        """

        def fn() -> None:
            ...

        fn.__signature__ = signature  # type:ignore[attr-defined]
        fn.__annotations__ = {p.name: p.annotation for p in signature.parameters.values()}
        return cls.from_fn(fn, signature_namespace)


def infer_request_encoding_from_parameter(param: ParsedParameter) -> RequestEncodingType | str:
    """Infer the request encoding type from a parsed type.

    Args:
        param: The parsed parameter to infer the request encoding type from.

    Returns:
        The inferred request encoding type.
    """
    if param.has_default and isinstance(param.default, BodyKwarg):
        return param.default.media_type
    if param.parsed_type.metadata:
        for item in param.parsed_type.metadata:
            if isinstance(item, BodyKwarg):
                return item.media_type
    return RequestEncodingType.JSON



from __future__ import annotations

from .abc import AbstractAsyncRepository, FilterTypes
from .exceptions import ConflictError, NotFoundError, RepositoryError

__all__ = (
    "AbstractAsyncRepository",
    "ConflictError",
    "FilterTypes",
    "NotFoundError",
    "RepositoryError",
)



from litestar.middleware.authentication import (
    AbstractAuthenticationMiddleware,
    AuthenticationResult,
)
from litestar.middleware.base import (
    AbstractMiddleware,
    DefineMiddleware,
    MiddlewareProtocol,
)

__all__ = (
    "AbstractAuthenticationMiddleware",
    "AbstractMiddleware",
    "AuthenticationResult",
    "DefineMiddleware",
    "MiddlewareProtocol",
)



from __future__ import annotations

import re
import typing as t
from collections import abc, defaultdict, deque
from typing import (
    AbstractSet,
    Any,
    DefaultDict,
    Deque,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import Annotated, NotRequired, Required, TypeGuard, get_args, get_origin

from litestar.types.builtin_types import UNION_TYPES, NoneType

__all__ = (
    "annotation_is_iterable_of_type",
    "get_origin_or_inner_type",
    "get_safe_generic_origin",
    "make_non_optional_union",
    "unwrap_annotation",
)


T = TypeVar("T")
UnionT = TypeVar("UnionT", bound="Union")

tuple_types_regex = re.compile(
    "^"
    + "|".join(
        [*[repr(x) for x in (List, Sequence, Iterable, Iterator, Tuple, Deque)], "tuple", "list", "collections.deque"]
    )
)

types_mapping = {
    AbstractSet: set,
    DefaultDict: defaultdict,
    Deque: deque,
    Dict: dict,
    FrozenSet: frozenset,
    List: list,
    Mapping: dict,
    MutableMapping: dict,
    MutableSequence: list,
    MutableSet: set,
    Sequence: list,
    Set: set,
    Tuple: tuple,
}

_safe_generic_origin_map = {
    set: t.AbstractSet,
    defaultdict: t.DefaultDict,
    deque: t.Deque,
    dict: t.Dict,
    frozenset: t.FrozenSet,
    list: t.List,
    tuple: t.Tuple,
    abc.Mapping: t.Mapping,
    abc.MutableMapping: t.MutableMapping,
    abc.MutableSequence: t.MutableSequence,
    abc.MutableSet: t.MutableSet,
    abc.Sequence: t.Sequence,
    abc.Set: t.AbstractSet,
    abc.Collection: t.Collection,
    abc.Container: t.Container,
    abc.ItemsView: t.ItemsView,
    abc.KeysView: t.KeysView,
    abc.MappingView: t.MappingView,
    abc.ValuesView: t.ValuesView,
    abc.Iterable: t.Iterable,
    abc.Iterator: t.Iterator,
    abc.Generator: t.Generator,
    abc.Reversible: t.Reversible,
    abc.Coroutine: t.Coroutine,
    abc.AsyncGenerator: t.AsyncGenerator,
    abc.AsyncIterable: t.AsyncIterable,
    abc.AsyncIterator: t.AsyncIterator,
    abc.Awaitable: t.Awaitable,
    **{union_t: t.Union for union_t in UNION_TYPES},  # type:ignore[misc]
}
"""A mapping of types to equivalent types that are safe to be used as generics across all Python versions.

This is necessary because occasionally we want to rebuild a generic outer type with different args, and types such as
``collections.abc.Mapping``, are not valid generic types in Python 3.8.
"""

wrapper_type_set = {Annotated, Required, NotRequired}
"""Types that always contain a wrapped type annotation as their first arg."""


def annotation_is_iterable_of_type(
    annotation: Any,
    type_value: type[T],
) -> TypeGuard[Iterable[T]]:
    """Determine if a given annotation is an iterable of the given type_value.

    Args:
        annotation: A type annotation.
        type_value: A type value.

    Returns:
        A type-guard boolean.
    """
    from litestar.utils.predicates import is_class_and_subclass

    if (args := get_args(annotation)) and (
        isinstance(annotation, (List, Sequence, Iterable, Iterator, Tuple, Deque, tuple, list, deque))  # type: ignore
        or tuple_types_regex.match(repr(annotation))
    ):
        return args[0] is type_value or isinstance(args[0], type_value) or is_class_and_subclass(args[0], type_value)
    return False


def make_non_optional_union(annotation: UnionT | None) -> UnionT:
    """Make a :data:`Union <typing.Union>` type that excludes ``NoneType``.

    Args:
        annotation: A type annotation.

    Returns:
        The union with all original members, except ``NoneType``.
    """
    args = tuple(tp for tp in get_args(annotation) if tp is not NoneType)
    return cast("UnionT", Union[args])  # pyright: ignore


def unwrap_union(annotation: Any) -> tuple[Any, ...]:
    """Unwrap a union type into a tuple of type arguments.

    Args:
        annotation: A union annotation.

    Returns:
        A tuple of annotations
    """
    from litestar.utils.predicates import is_optional_union, is_union

    args: list[Any] = []

    for arg in get_args(annotation):
        arg_value = get_origin_or_inner_type(arg) or arg
        if is_optional_union(arg_value) or is_union(arg_value):
            args.extend(unwrap_union(arg_value))
        else:
            args.append(arg_value)

    return tuple(args)


def unwrap_annotation(annotation: Any) -> tuple[Any, tuple[Any, ...], set[Any]]:
    """Remove "wrapper" annotation types, such as ``Annotated``, ``Required``, and ``NotRequired``.

    Note:
        ``annotation`` should have been retrieved from :func:`get_type_hints()` with ``include_extras=True``. This
        ensures that any nested ``Annotated`` types are flattened according to the PEP 593 specification.

    Args:
        annotation: A type annotation.

    Returns:
        A tuple of the unwrapped annotation and any ``Annotated`` metadata, and a set of any wrapper types encountered.
    """
    origin = get_origin(annotation)
    wrappers = set()
    metadata = []
    while origin in wrapper_type_set:
        wrappers.add(origin)
        annotation, *meta = get_args(annotation)
        metadata.extend(meta)
        origin = get_origin(annotation)
    return annotation, tuple(metadata), wrappers


def get_origin_or_inner_type(annotation: Any) -> Any:
    """Get origin or unwrap it. Returns None for non-generic types.

    Args:
        annotation: A type annotation.

    Returns:
        Any type.
    """
    origin = get_origin(annotation)
    if origin in wrapper_type_set:
        inner, _, _ = unwrap_annotation(annotation)
        # we need to recursively call here 'get_origin_or_inner_type' because we might be dealing with a generic type alias
        # e.g. Annotated[dict[str, list[int]]
        origin = get_origin_or_inner_type(inner)
    return types_mapping.get(origin, origin)


def get_safe_generic_origin(origin_type: Any) -> Any:
    """Get a type that is safe to use as a generic type across all supported Python versions.

    Args:
        origin_type: A type - would be the return value of :func:`get_origin()`.

    Returns:
        The ``typing`` module equivalent of the given type, if it exists. Otherwise, the original type is returned.
    """
    return _safe_generic_origin_map.get(origin_type, origin_type)



from typing import Awaitable, Literal, Optional, Sequence, TypeVar, Union

T = TypeVar("T")

OptionalSequence = Optional[Sequence[T]]
"""Types 'T' as union of Sequence[T] and None."""
SyncOrAsyncUnion = Union[T, Awaitable[T]]
"""Types 'T' as a union of T and awaitable T."""
AnyIOBackend = Literal["asyncio", "trio"]
"""Anyio backend names."""



from __future__ import annotations

from .abc import AbstractDTOBackend
from .msgspec.backend import MsgspecDTOBackend
from .pydantic.backend import PydanticDTOBackend

__all__ = ("AbstractDTOBackend", "MsgspecDTOBackend", "PydanticDTOBackend")



"""Collection filter datastructures."""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, Literal, TypeVar

if TYPE_CHECKING:
    from collections import abc
    from datetime import datetime

T = TypeVar("T")

__all__ = ["BeforeAfter", "CollectionFilter", "LimitOffset", "OrderBy", "SearchFilter"]


@dataclass
class BeforeAfter:
    """Data required to filter a query on a ``datetime`` column."""

    field_name: str
    """Name of the model attribute to filter on."""
    before: datetime | None
    """Filter results where field earlier than this."""
    after: datetime | None
    """Filter results where field later than this."""


@dataclass
class CollectionFilter(Generic[T]):
    """Data required to construct a ``WHERE ... IN (...)`` clause."""

    field_name: str
    """Name of the model attribute to filter on."""
    values: abc.Collection[T]
    """Values for ``IN`` clause."""


@dataclass
class LimitOffset:
    """Data required to add limit/offset filtering to a query."""

    limit: int
    """Value for ``LIMIT`` clause of query."""
    offset: int
    """Value for ``OFFSET`` clause of query."""


@dataclass
class OrderBy(Generic[T]):
    """Data required to construct a ``ORDER BY ...`` clause."""

    field_name: str
    """Name of the model attribute to sort on."""
    sort_order: Literal["asc", "desc"]
    """Sort ascending or descending"""


@dataclass
class SearchFilter(Generic[T]):
    """Data required to construct a ``WHERE field_name LIKE '%' || :value || '%'`` clause."""

    field_name: str
    """Name of the model attribute to sort on."""
    value: str
    """Values for ``LIKE`` clause."""
    ignore_case: bool | None = False
    """Should the search be case insensitive."""



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from litestar.exceptions import ImproperlyConfiguredException
from litestar.file_system import BaseLocalFileSystem
from litestar.handlers import asgi
from litestar.static_files.base import StaticFiles
from litestar.utils import normalize_path

__all__ = ("StaticFilesConfig",)


if TYPE_CHECKING:
    from litestar.handlers.asgi_handlers import ASGIRouteHandler
    from litestar.types import ExceptionHandlersMap, Guard, PathType


@dataclass
class StaticFilesConfig:
    """Configuration for static file service.

    To enable static files, pass an instance of this class to the :class:`Litestar <litestar.app.Litestar>` constructor using
    the 'static_files_config' key.
    """

    path: str
    """Path to serve static files from.

    Note that the path cannot contain path parameters.
    """
    directories: list[PathType]
    """A list of directories to serve files from."""
    html_mode: bool = False
    """Flag dictating whether serving html.

    If true, the default file will be 'index.html'.
    """
    name: str | None = None
    """An optional string identifying the static files handler."""
    file_system: Any = BaseLocalFileSystem()
    """The file_system spec to use for serving files.

    Notes:
        - A file_system is a class that adheres to the
            :class:`FileSystemProtocol <litestar.types.FileSystemProtocol>`.
        - You can use any of the file systems exported from the
            [fsspec](https://filesystem-spec.readthedocs.io/en/latest/) library for this purpose.
    """
    opt: dict[str, Any] | None = None
    """A string key dictionary of arbitrary values that will be added to the static files handler."""
    guards: list[Guard] | None = None
    """A list of :class:`Guard <litestar.types.Guard>` callables."""
    exception_handlers: ExceptionHandlersMap | None = None
    """A dictionary that maps handler functions to status codes and/or exception types."""
    send_as_attachment: bool = False
    """Whether to send the file as an attachment."""

    def __post_init__(self) -> None:
        if not self.path:
            raise ImproperlyConfiguredException("path must be a non-zero length string,")

        if not self.directories or not any(bool(d) for d in self.directories):
            raise ImproperlyConfiguredException("directories must include at least one path.")

        if "{" in self.path:
            raise ImproperlyConfiguredException("path parameters are not supported for static files")

        if not (
            callable(getattr(self.file_system, "info", None)) and callable(getattr(self.file_system, "open", None))
        ):
            raise ImproperlyConfiguredException("file_system must adhere to the FileSystemProtocol type")

        self.path = normalize_path(self.path)

    def to_static_files_app(self) -> ASGIRouteHandler:
        """Return an ASGI app serving static files based on the config.

        Returns:
            :class:`StaticFiles <litestar.static_files.StaticFiles>`
        """
        static_files = StaticFiles(
            is_html_mode=self.html_mode,
            directories=self.directories,
            file_system=self.file_system,
            send_as_attachment=self.send_as_attachment,
        )
        return asgi(
            path=self.path,
            name=self.name,
            is_static=True,
            opt=self.opt,
            guards=self.guards,
            exception_handlers=self.exception_handlers,
        )(static_files)



from __future__ import annotations

from abc import ABC, abstractmethod
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Mapping, cast

from msgspec.json import Encoder as JsonEncoder

from litestar._signature import create_signature_model
from litestar.dto.interface import HandlerContext
from litestar.exceptions import ImproperlyConfiguredException
from litestar.serialization import default_serializer
from litestar.types import (
    AnyCallable,
    Dependencies,
    Empty,
    EmptyType,
    ExceptionHandler,
    Guard,
    Middleware,
    SyncOrAsyncUnion,
    TypeEncodersMap,
)
from litestar.types.builtin_types import NoneType
from litestar.utils import AsyncCallable
from litestar.utils.signature import ParsedSignature

from . import _utils
from .route_handler import WebsocketRouteHandler

if TYPE_CHECKING:
    from litestar import Litestar, WebSocket
    from litestar.dto.interface import DTOInterface
    from litestar.types.asgi_types import WebSocketMode

__all__ = (
    "WebsocketListener",
    "websocket_listener",
)


class websocket_listener(WebsocketRouteHandler):
    """A websocket listener that automatically accepts a connection, handles disconnects,
    invokes a callback function every time new data is received and sends any data
    returned
    """

    __slots__ = (
        "_on_accept",
        "_on_disconnect",
        "_pass_socket",
        "_receive_mode",
        "_send_mode",
        "_listener_context",
    )

    def __init__(
        self,
        path: str | None | list[str] | None = None,
        *,
        dependencies: Dependencies | None = None,
        dto: type[DTOInterface] | None | EmptyType = Empty,
        exception_handlers: dict[int | type[Exception], ExceptionHandler] | None = None,
        guards: list[Guard] | None = None,
        middleware: list[Middleware] | None = None,
        receive_mode: WebSocketMode = "text",
        send_mode: WebSocketMode = "text",
        name: str | None = None,
        on_accept: Callable[[WebSocket], SyncOrAsyncUnion[None]] | None = None,
        on_disconnect: Callable[[WebSocket], SyncOrAsyncUnion[None]] | None = None,
        opt: dict[str, Any] | None = None,
        return_dto: type[DTOInterface] | None | EmptyType = Empty,
        signature_namespace: Mapping[str, Any] | None = None,
        type_encoders: TypeEncodersMap | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize ``WebsocketRouteHandler``

        Args:
            path: A path fragment for the route handler function or a sequence of path fragments. If not given defaults
                to ``/``
            dependencies: A string keyed mapping of dependency :class:`Provider <.di.Provide>` instances.
            dto: :class:`DTOInterface <.dto.interface.DTOInterface>` to use for (de)serializing and
                validation of request data.
            exception_handlers: A mapping of status codes and/or exception types to handler functions.
            guards: A sequence of :class:`Guard <.types.Guard>` callables.
            middleware: A sequence of :class:`Middleware <.types.Middleware>`.
            receive_mode: Websocket mode to receive data in, either `text` or `binary`.
            send_mode: Websocket mode to receive data in, either `text` or `binary`.
            name: A string identifying the route handler.
            on_accept: Callback invoked after a connection has been accepted, receiving the
                :class:`WebSocket <.connection.WebSocket>` instance as its only argument
            on_disconnect: Callback invoked after a connection has been closed, receiving the
                :class:`WebSocket <.connection.WebSocket>` instance as its only argument
            opt: A string keyed mapping of arbitrary values that can be accessed in :class:`Guards <.types.Guard>` or
                wherever you have access to :class:`Request <.connection.Request>` or
                :class:`ASGI Scope <.types.Scope>`.
            return_dto: :class:`DTOInterface <.dto.interface.DTOInterface>` to use for serializing
                outbound response data.
            signature_namespace: A mapping of names to types for use in forward reference resolution during signature
                modelling.
            type_encoders: A mapping of types to callables that transform them into types supported for serialization.
            **kwargs: Any additional kwarg - will be set in the opt dictionary.
        """
        self._listener_context = _utils.ListenerContext()
        self._receive_mode: WebSocketMode = receive_mode
        self._send_mode: WebSocketMode = send_mode
        self._on_accept = AsyncCallable(on_accept) if on_accept else None
        self._on_disconnect = AsyncCallable(on_disconnect) if on_disconnect else None
        self.type_encoders = type_encoders

        super().__init__(
            path=path,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            guards=guards,
            middleware=middleware,
            name=name,
            opt=opt,
            signature_namespace=signature_namespace,
            **kwargs,
        )

        # need to be assigned after the super() call
        self.dto = dto
        self.return_dto = return_dto

    def _validate_handler_function(self) -> None:
        """Validate the route handler function once it's set by inspecting its return annotations."""
        # since none of the validation rules of WebsocketRouteHandler apply here, this is let empty. Validation of the
        # user supplied method happens at init time of this handler instead in __call__

    def _init_handler_dtos(self) -> None:
        """Initialize the data and return DTOs for the handler."""
        if dto := self.resolve_dto():
            data_parameter = self._listener_context.listener_callback_signature.parameters["data"]
            dto.on_registration(HandlerContext("data", str(self), data_parameter.parsed_type))

        if return_dto := self.resolve_return_dto():
            return_type = self._listener_context.listener_callback_signature.return_type
            return_dto.on_registration(HandlerContext("return", str(self), return_type))

    def __call__(self, listener_callback: AnyCallable) -> websocket_listener:
        self._listener_context.listener_callback = listener_callback
        self._listener_context.handler_function = handler_function = _utils.create_handler_function(
            listener_context=self._listener_context,
            on_accept=self._on_accept,
            on_disconnect=self._on_disconnect,
        )
        return super().__call__(handler_function)

    def on_registration(self, app: Litestar) -> None:
        self._set_listener_context()
        super().on_registration(app)

    def _create_signature_model(self, app: Litestar) -> None:
        """Create signature model for handler function."""
        if not self.signature_model:
            new_signature = _utils.create_handler_signature(
                self._listener_context.listener_callback_signature.original_signature
            )
            self.signature_model = create_signature_model(
                dependency_name_set=self.dependency_name_set,
                fn=cast("AnyCallable", self.fn.value),
                preferred_validation_backend=app.preferred_validation_backend,
                parsed_signature=ParsedSignature.from_signature(new_signature, self.resolve_signature_namespace()),
            )

    def _set_listener_context(self) -> None:
        listener_context = self._listener_context
        listener_context.listener_callback_signature = listener_callback_signature = ParsedSignature.from_fn(
            listener_context.listener_callback, self.resolve_signature_namespace()
        )

        if "data" not in listener_callback_signature.parameters:
            raise ImproperlyConfiguredException("Websocket listeners must accept a 'data' parameter")

        for param in ("request", "body"):
            if param in listener_callback_signature.parameters:
                raise ImproperlyConfiguredException(f"The {param} kwarg is not supported with websocket listeners")

        listener_context.can_send_data = not listener_callback_signature.return_type.is_subclass_of(NoneType)
        listener_context.pass_socket = "socket" in listener_callback_signature.parameters
        listener_context.resolved_data_dto = resolved_data_dto = self.resolve_dto()
        listener_context.resolved_return_dto = resolved_return_dto = self.resolve_return_dto()
        listener_context.handle_receive = _utils.create_handle_receive(
            resolved_data_dto, self._receive_mode, listener_callback_signature.parameters["data"].annotation
        )
        should_encode_to_json = not (
            listener_callback_signature.return_type.is_subclass_of((str, bytes))
            or (
                listener_callback_signature.return_type.is_optional
                and listener_callback_signature.return_type.has_inner_subclass_of((str, bytes))
            )
        )
        json_encoder = JsonEncoder(enc_hook=partial(default_serializer, type_encoders=self.resolve_type_encoders()))
        listener_context.handle_send = _utils.create_handle_send(
            resolved_return_dto, json_encoder, should_encode_to_json, self._send_mode
        )


class WebsocketListener(ABC):
    path: str | None | list[str] | None = None
    """A path fragment for the route handler function or a sequence of path fragments. If not given defaults to ``/``"""
    dependencies: Dependencies | None = None
    """A string keyed mapping of dependency :class:`Provider <.di.Provide>` instances."""
    dto: type[DTOInterface] | None | EmptyType = Empty
    """:class:`DTOInterface <.dto.interface.DTOInterface>` to use for (de)serializing and validation of request data"""
    exception_handlers: dict[int | type[Exception], ExceptionHandler] | None = None
    """A mapping of status codes and/or exception types to handler functions."""
    guards: list[Guard] | None = None
    """A sequence of :class:`Guard <.types.Guard>` callables."""
    middleware: list[Middleware] | None = None
    """A sequence of :class:`Middleware <.types.Middleware>`."""
    receive_mode: WebSocketMode = "text"
    """Websocket mode to receive data in, either `text` or `binary`."""
    send_mode: WebSocketMode = "text"
    """Websocket mode to send data in, either `text` or `binary`."""
    name: str | None = None
    """A string identifying the route handler."""
    opt: dict[str, Any] | None = None
    """
    A string keyed mapping of arbitrary values that can be accessed in :class:`Guards <.types.Guard>` or wherever you
    have access to :class:`Request <.connection.Request>` or :class:`ASGI Scope <.types.Scope>`.
    """
    return_dto: type[DTOInterface] | None | EmptyType = Empty
    """:class:`DTOInterface <.dto.interface.DTOInterface>` to use for serializing outbound response data."""
    signature_namespace: Mapping[str, Any] | None = None
    """
    A mapping of names to types for use in forward reference resolution during signature modelling.
    """
    type_encoders: TypeEncodersMap | None = None
    """
    type_encoders: A mapping of types to callables that transform them into types supported for serialization.
    """

    def __init__(self) -> None:
        self._handler = websocket_listener(
            dependencies=self.dependencies,
            dto=self.dto,
            exception_handlers=self.exception_handlers,
            guards=self.guards,
            middleware=self.middleware,
            send_mode=self.send_mode,
            receive_mode=self.receive_mode,
            name=self.name,
            on_accept=self.on_accept,
            on_disconnect=self.on_disconnect,
            opt=self.opt,
            path=self.path,
            return_dto=self.return_dto,
            signature_namespace=self.signature_namespace,
            type_encoders=self.type_encoders,
        )(self.on_receive)

    def on_accept(self, socket: WebSocket) -> SyncOrAsyncUnion[None]:  # noqa: B027
        """Called after a WebSocket connection has been accepted"""

    @abstractmethod
    def on_receive(self, *args: Any, **kwargs: Any) -> Any:
        """Called after data has been received from the WebSocket.

        This should take a ``data`` argument, receiving the processed WebSocket data,
        and can additionally include handler dependencies such as ``state``, or other
        regular dependencies.

        Data returned from this function will be serialized and sent via the socket
        according to handler configuration.
        """
        raise NotImplementedError

    def on_disconnect(self, socket: WebSocket) -> SyncOrAsyncUnion[None]:  # noqa: B027
        """Called after a WebSocket connection has been disconnected"""



from __future__ import annotations

from collections.abc import Collection as CollectionsCollection
from typing import TYPE_CHECKING, NewType, TypeVar, cast
from uuid import uuid4

from msgspec import Struct, from_builtins

from litestar.dto.factory.backends.abc import AbstractDTOBackend, BackendContext
from litestar.serialization import decode_media_type

from .utils import _build_data_from_struct, _build_struct_from_model, _create_struct_for_field_definitions

if TYPE_CHECKING:
    from typing import Any, Collection

    from litestar.dto.interface import ConnectionContext
    from litestar.types.serialization import LitestarEncodableType

__all__ = ("MsgspecDTOBackend",)


MsgspecField = NewType("MsgspecField", type)
T = TypeVar("T")


class MsgspecDTOBackend(AbstractDTOBackend[Struct]):
    __slots__ = ()

    def create_data_container_type(self, context: BackendContext) -> type[Struct]:
        return _create_struct_for_field_definitions(str(uuid4()), context.field_definitions)

    def parse_raw(self, raw: bytes, connection_context: ConnectionContext) -> Struct | Collection[Struct]:
        return decode_media_type(  # type:ignore[no-any-return]
            raw, connection_context.request_encoding_type, type_=self.annotation
        )

    def populate_data_from_builtins(self, data: Any) -> Any:
        parsed_data = cast("Struct | Collection[Struct]", from_builtins(data, self.annotation))
        return _build_data_from_struct(
            self.context.model_type, parsed_data, self.context.field_definitions, self.context.reverse_name_map
        )

    def populate_data_from_raw(self, raw: bytes, connection_context: ConnectionContext) -> T | Collection[T]:
        parsed_data = self.parse_raw(raw, connection_context)
        return _build_data_from_struct(
            self.context.model_type, parsed_data, self.context.field_definitions, self.context.reverse_name_map
        )

    def encode_data(self, data: Any, connection_context: ConnectionContext) -> LitestarEncodableType:
        if isinstance(data, CollectionsCollection):
            return self.context.parsed_type.origin(  # type:ignore[no-any-return]
                _build_struct_from_model(datum, self.data_container_type, self.context.reverse_name_map)
                for datum in data  # pyright:ignore
            )
        return _build_struct_from_model(data, self.data_container_type, self.context.reverse_name_map)



from __future__ import annotations

from math import inf
from typing import TYPE_CHECKING, Generic, Optional, TypeVar, cast

from anyio import create_memory_object_stream
from anyio.streams.stapled import StapledObjectStream

from litestar.testing.client.base import BaseTestClient

if TYPE_CHECKING:
    from litestar.types import (
        LifeSpanReceiveMessage,  # noqa: F401
        LifeSpanSendMessage,
        LifeSpanShutdownEvent,
        LifeSpanStartupEvent,
    )

T = TypeVar("T", bound=BaseTestClient)


class LifeSpanHandler(Generic[T]):
    __slots__ = "stream_send", "stream_receive", "client", "task"

    def __init__(self, client: T):
        self.client = client
        self.stream_send = StapledObjectStream[Optional["LifeSpanSendMessage"]](*create_memory_object_stream(inf))
        self.stream_receive = StapledObjectStream["LifeSpanReceiveMessage"](*create_memory_object_stream(inf))

        with self.client.portal() as portal:
            self.task = portal.start_task_soon(self.lifespan)
            portal.call(self.wait_startup)

    async def receive(self) -> LifeSpanSendMessage:
        message = await self.stream_send.receive()
        if message is None:
            self.task.result()
        return cast("LifeSpanSendMessage", message)

    async def wait_startup(self) -> None:
        event: LifeSpanStartupEvent = {"type": "lifespan.startup"}
        await self.stream_receive.send(event)

        message = await self.receive()
        if message["type"] not in (
            "lifespan.startup.complete",
            "lifespan.startup.failed",
        ):
            raise AssertionError
        if message["type"] == "lifespan.startup.failed":
            await self.receive()

    async def wait_shutdown(self) -> None:
        async with self.stream_send:
            lifespan_shutdown_event: LifeSpanShutdownEvent = {"type": "lifespan.shutdown"}
            await self.stream_receive.send(lifespan_shutdown_event)

            message = await self.receive()
            if message["type"] not in (
                "lifespan.shutdown.complete",
                "lifespan.shutdown.failed",
            ):
                raise AssertionError
            if message["type"] == "lifespan.shutdown.failed":
                await self.receive()

    async def lifespan(self) -> None:
        scope = {"type": "lifespan"}
        try:
            await self.client.app(scope, self.stream_receive.receive, self.stream_send.send)
        finally:
            await self.stream_send.send(None)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, TypedDict, TypeVar, Union, runtime_checkable

from pydantic import BaseModel

from litestar.types.protocols import DataclassProtocol

if TYPE_CHECKING:
    from typing_extensions import TypeGuard

    from litestar.config.app import AppConfig
    from litestar.dto.interface import DTOInterface
    from litestar.openapi.spec import Schema
    from litestar.utils.signature import ParsedType

__all__ = ("SerializationPluginProtocol", "InitPluginProtocol", "OpenAPISchemaPluginProtocol", "PluginProtocol")

ModelT = TypeVar("ModelT")
DataContainerT = TypeVar("DataContainerT", bound=Union[BaseModel, DataclassProtocol, TypedDict])  # type: ignore[valid-type]


@runtime_checkable
class InitPluginProtocol(Protocol):
    """Protocol used to define plugins that affect the application's init process."""

    __slots__ = ()

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        """Receive the :class:`AppConfig<.config.app.AppConfig>` instance after `on_app_init` hooks have been called.

        Examples:
            .. code-block: python
                from litestar import Litestar, get
                from litestar.di import Provide
                from litestar.plugins import InitPluginProtocol

                def get_name() -> str:
                    return "world"

                @get("/my-path")
                def my_route_handler(name: str) -> dict[str, str]:
                    return {"hello": name}

                class MyPlugin(InitPluginProtocol):
                    def on_app_init(self, app_config: AppConfig) -> AppConfig:
                        app_config.dependencies["name"] = Provide(get_name)
                        app_config.route_handlers.append(my_route_handler)
                        return app_config

                app = Litestar(plugins=[MyPlugin()])

        Args:
            app_config: The :class:`AppConfig <litestar.config.app.AppConfig>` instance.

        Returns:
            The app config object.
        """
        return app_config  # pragma: no cover


@runtime_checkable
class SerializationPluginProtocol(Protocol):
    """Protocol used to define a serialization plugin for DTOs."""

    __slots__ = ()

    @staticmethod
    def supports_type(parsed_type: ParsedType) -> bool:
        """Given a value of indeterminate type, determine if this value is supported by the plugin.

        Args:
            parsed_type: A parsed type.

        Returns:
            Whether the type is supported by the plugin.
        """
        raise NotImplementedError()

    def create_dto_for_type(self, parsed_type: ParsedType) -> type[DTOInterface]:
        """Given a parsed type, create a DTO class.

        Args:
            parsed_type: A parsed type.

        Returns:
            A DTO class.
        """
        raise NotImplementedError()


@runtime_checkable
class OpenAPISchemaPluginProtocol(Protocol[ModelT]):
    """Plugin to extend the support of OpenAPI schema generation for non-library types."""

    __slots__ = ()

    @staticmethod
    def is_plugin_supported_type(value: Any) -> TypeGuard[ModelT]:
        """Given a value of indeterminate type, determine if this value is supported by the plugin.

        Args:
            value: An arbitrary value.

        Returns:
            A typeguard dictating whether the value is supported by the plugin.
        """
        raise NotImplementedError()

    def to_openapi_schema(self, model_class: type[ModelT]) -> Schema:
        """Given a model class, transform it into an OpenAPI schema class.

        Args:
            model_class: A model class.

        Returns:
            An :class:`OpenAPI <litestar.openapi.spec.schema.Schema>` instance.
        """
        raise NotImplementedError()


PluginProtocol = Union[SerializationPluginProtocol, InitPluginProtocol, OpenAPISchemaPluginProtocol]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Protocol, TypeVar, runtime_checkable

from typing_extensions import TypedDict

__all__ = (
    "TemplateContext",
    "TemplateEngineProtocol",
    "TemplateProtocol",
    "csrf_token",
    "url_for",
    "url_for_static_asset",
)


if TYPE_CHECKING:
    from pydantic import DirectoryPath

    from litestar.connection import Request


class TemplateContext(TypedDict):
    """Dictionary representing a template context."""

    request: Request[Any, Any, Any]
    csrf_input: str


def url_for(context: TemplateContext, route_name: str, **path_parameters: Any) -> str:
    """Wrap :func:`route_reverse <litestar.app.route_reverse>` to be used in templates.

    Args:
        context: The template context.
        route_name: The name of the route handler.
        **path_parameters: Actual values for path parameters in the route.

    Raises:
        NoRouteMatchFoundException: If ``route_name`` does not exist, path parameters are missing in **path_parameters or have wrong type.

    Returns:
        A fully formatted url path.
    """
    return context["request"].app.route_reverse(route_name, **path_parameters)


def csrf_token(context: TemplateContext) -> str:
    """Set a CSRF token on the template.

    Notes:
        - to use this function make sure to pass an instance of :ref:`CSRFConfig <litestar.config.csrf_config.CSRFConfig>` to
        the :class:`Litestar <litestar.app.Litestar>` constructor.

    Args:
        context: The template context.

    Returns:
        A CSRF token if the app level ``csrf_config`` is set, otherwise an empty string.
    """
    return context["request"].scope.get("_csrf_token", "")  # type: ignore


def url_for_static_asset(context: TemplateContext, name: str, file_path: str) -> str:
    """Wrap :meth:`url_for_static_asset <litestar.app.url_for_static_asset>` to be used in templates.

    Args:
        context: The template context object.
        name: A static handler unique name.
        file_path: a string containing path to an asset.

    Raises:
        NoRouteMatchFoundException: If static files handler with ``name`` does not exist.

    Returns:
        A url path to the asset.
    """
    return context["request"].app.url_for_static_asset(name, file_path)


class TemplateProtocol(Protocol):  # pragma: no cover
    """Protocol Defining a ``Template``.

    Template is a class that has a render method which renders the template into a string.
    """

    def render(self, *args: Any, **kwargs: Any) -> str:
        """Return the rendered template as a string.

        Args:
            *args: Positional arguments passed to the TemplateEngine
            **kwargs: A string keyed mapping of values passed to the TemplateEngine

        Returns:
            The rendered template string
        """
        ...


T_co = TypeVar("T_co", bound=TemplateProtocol, covariant=True)


@runtime_checkable
class TemplateEngineProtocol(Protocol[T_co]):  # pragma: no cover
    """Protocol for template engines."""

    def __init__(self, directory: DirectoryPath | list[DirectoryPath]) -> None:
        """Initialize the template engine with a directory.

        Args:
            directory: Direct path or list of directory paths from which to serve templates.
        """
        ...

    def get_template(self, template_name: str) -> T_co:
        """Retrieve a template by matching its name (dotted path) with files in the directory or directories provided.

        Args:
            template_name: A dotted path

        Returns:
            Template instance

        Raises:
            TemplateNotFoundException: if no template is found.
        """
        ...

    def register_template_callable(self, key: str, template_callable: Callable[[dict[str, Any]], Any]) -> None:
        """Register a callable on the template engine.

        Args:
            key: The callable key, i.e. the value to use inside the template to call the callable.
            template_callable: A callable to register.

        Returns:
            None
        """



from __future__ import annotations

from functools import lru_cache
from inspect import isawaitable
from typing import TYPE_CHECKING, Any, Sequence, cast

from litestar.dto.interface import ConnectionContext
from litestar.enums import HttpMethod
from litestar.exceptions import ValidationException
from litestar.status_codes import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT

if TYPE_CHECKING:
    from litestar.app import Litestar
    from litestar.background_tasks import BackgroundTask, BackgroundTasks
    from litestar.connection import Request
    from litestar.datastructures import Cookie, ResponseHeader
    from litestar.dto.interface import DTOInterface
    from litestar.response import Response
    from litestar.response_containers import ResponseContainer
    from litestar.types import (
        AfterRequestHookHandler,
        ASGIApp,
        AsyncAnyCallable,
        Method,
        ResponseType,
        TypeEncodersMap,
    )

__all__ = (
    "create_data_handler",
    "create_generic_asgi_response_handler",
    "create_response_container_handler",
    "create_response_handler",
    "filter_cookies",
    "get_default_status_code",
    "normalize_headers",
    "normalize_http_method",
)


def create_data_handler(
    after_request: AfterRequestHookHandler | None,
    background: BackgroundTask | BackgroundTasks | None,
    cookies: frozenset[Cookie],
    headers: frozenset[ResponseHeader],
    media_type: str,
    response_class: ResponseType,
    status_code: int,
    type_encoders: TypeEncodersMap | None,
) -> AsyncAnyCallable:
    """Create a handler function for arbitrary data.

    Args:
        after_request: An after request handler.
        background: A background task or background tasks.
        cookies: A set of pre-defined cookies.
        headers: A set of response headers.
        media_type: The response media type.
        response_class: The response class to use.
        status_code: The response status code.
        type_encoders: A mapping of types to encoder functions.

    Returns:
        A handler function.

    """
    normalized_headers = [
        (name.lower().encode("latin-1"), value.encode("latin-1")) for name, value in normalize_headers(headers).items()
    ]
    cookie_headers = [cookie.to_encoded_header() for cookie in cookies if not cookie.documentation_only]
    raw_headers = [*normalized_headers, *cookie_headers]

    async def create_response(data: Any) -> ASGIApp:
        response = response_class(
            background=background,
            content=data,
            media_type=media_type,
            status_code=status_code,
            type_encoders=type_encoders,
        )
        response.raw_headers = raw_headers

        if after_request:
            return await after_request(response)  # type: ignore

        return response

    async def handler(
        data: Any,
        return_dto: type[DTOInterface] | None,
        request: Request[Any, Any, Any],
        **kwargs: Any,
    ) -> ASGIApp:
        if isawaitable(data):
            data = await data

        if return_dto:
            ctx = ConnectionContext.from_connection(request)
            data = return_dto(ctx).data_to_encodable_type(data)

        return await create_response(data=data)

    return handler


def filter_cookies(local_cookies: frozenset[Cookie], layered_cookies: frozenset[Cookie]) -> list[Cookie]:
    """Given two sets of cookies, return a unique list of cookies, that are not marked as documentation_only.

    Args:
        local_cookies: Cookies returned from the local scope.
        layered_cookies: Cookies returned from the layers.

    Returns:
        A unified list of cookies
    """
    return [cookie for cookie in {*local_cookies, *layered_cookies} if not cookie.documentation_only]


def create_generic_asgi_response_handler(
    after_request: AfterRequestHookHandler | None,
    cookies: frozenset[Cookie],
) -> AsyncAnyCallable:
    """Create a handler function for Responses.

    Args:
        after_request: An after request handler.
        cookies: A set of pre-defined cookies.

    Returns:
        A handler function.
    """

    async def handler(data: ASGIApp, **kwargs: Any) -> ASGIApp:
        if hasattr(data, "set_cookie"):
            for cookie in cookies:
                data.set_cookie(**cookie.dict)
        return await after_request(data) if after_request else data  # type: ignore

    return handler


@lru_cache(1024)
def normalize_headers(headers: frozenset[ResponseHeader]) -> dict[str, str]:
    """Given a dictionary of ResponseHeader, filter them and return a dictionary of values.

    Args:
        headers: A dictionary of :class:`ResponseHeader <litestar.datastructures.ResponseHeader>` values

    Returns:
        A string keyed dictionary of normalized values
    """
    return {
        header.name: cast("str", header.value)  # we know value to be a string at this point because we validate it
        # that it's not None when initializing a header with documentation_only=True
        for header in headers
        if not header.documentation_only
    }


def create_response_container_handler(
    after_request: AfterRequestHookHandler | None,
    cookies: frozenset[Cookie],
    headers: frozenset[ResponseHeader],
    media_type: str,
    status_code: int,
) -> AsyncAnyCallable:
    """Create a handler function for ResponseContainers.

    Args:
        after_request: An after request handler.
        cookies: A set of pre-defined cookies.
        headers: A set of response headers.
        media_type: The response media type.
        status_code: The response status code.

    Returns:
        A handler function.
    """
    normalized_headers = normalize_headers(headers)

    async def handler(data: ResponseContainer, app: Litestar, request: Request, **kwargs: Any) -> ASGIApp:
        response = data.to_response(
            app=app,
            headers={**normalized_headers, **data.headers},
            status_code=status_code,
            media_type=data.media_type or media_type,
            request=request,
        )
        response.cookies = filter_cookies(frozenset(data.cookies), cookies)
        return await after_request(response) if after_request else response  # type: ignore

    return handler


def create_response_handler(
    after_request: AfterRequestHookHandler | None,
    cookies: frozenset[Cookie],
) -> AsyncAnyCallable:
    """Create a handler function for Litestar Responses.

    Args:
        after_request: An after request handler.
        cookies: A set of pre-defined cookies.

    Returns:
        A handler function.
    """

    async def handler(data: Response, **kwargs: Any) -> ASGIApp:
        data.cookies = filter_cookies(frozenset(data.cookies), cookies)
        return await after_request(data) if after_request else data  # type: ignore

    return handler


def normalize_http_method(http_methods: HttpMethod | Method | Sequence[HttpMethod | Method]) -> set[Method]:
    """Normalize HTTP method(s) into a set of upper-case method names.

    Args:
        http_methods: A value for http method.

    Returns:
        A normalized set of http methods.
    """
    output: set[str] = set()

    if isinstance(http_methods, str):
        http_methods = [http_methods]  # pyright: ignore

    for method in http_methods:
        method_name = method.value.upper() if isinstance(method, HttpMethod) else method.upper()
        if method_name not in HTTP_METHOD_NAMES:
            raise ValidationException(f"Invalid HTTP method: {method_name}")
        output.add(method_name)

    return cast("set[Method]", output)


def get_default_status_code(http_methods: set[Method]) -> int:
    """Return the default status code for a given set of HTTP methods.

    Args:
        http_methods: A set of method strings

    Returns:
        A status code
    """
    if HttpMethod.POST in http_methods:
        return HTTP_201_CREATED
    if HttpMethod.DELETE in http_methods:
        return HTTP_204_NO_CONTENT
    return HTTP_200_OK


HTTP_METHOD_NAMES = {m.value for m in HttpMethod}



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from typing import Type

    from typing_extensions import TypeAlias, _TypedDictMeta  # type: ignore


if sys.version_info >= (3, 10):
    from types import UnionType

    UNION_TYPES = {UnionType, Union}
else:  # pragma: no cover
    UNION_TYPES = {Union}

NoneType: type[None] = type(None)
TypedDictClass: TypeAlias = "Type[_TypedDictMeta]"



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Callable, Coroutine, cast

from litestar.dto.interface import ConnectionContext
from litestar.exceptions import WebSocketDisconnect
from litestar.serialization import decode_json
from litestar.utils import AsyncCallable

if TYPE_CHECKING:
    from msgspec.json import Encoder as JsonEncoder

    from litestar import WebSocket
    from litestar.dto.interface import DTOInterface
    from litestar.types import AnyCallable, TypeEncodersMap
    from litestar.types.asgi_types import WebSocketMode
    from litestar.utils.signature import ParsedSignature


class ListenerContext:
    __slots__ = (
        "can_send_data",
        "handler_function",
        "handle_receive",
        "handle_send",
        "json_encoder",
        "listener_callback",
        "listener_callback_signature",
        "pass_socket",
        "resolved_data_dto",
        "resolved_return_dto",
    )

    can_send_data: bool
    handler_function: AnyCallable
    handle_receive: Callable[[WebSocket, DTOInterface | None], Any]
    handle_send: Callable[[WebSocket, Any, DTOInterface | None], Coroutine[None, None, None]]
    json_encoder: JsonEncoder
    listener_callback: AnyCallable
    listener_callback_signature: ParsedSignature
    pass_socket: bool
    resolved_data_dto: type[DTOInterface] | None
    resolved_return_dto: type[DTOInterface] | None
    resolved_type_encoders: TypeEncodersMap


def create_handle_receive(
    resolved_data_dto: type[DTOInterface] | None,
    receive_mode: WebSocketMode,
    wants_receive_type: type,
) -> Callable[[WebSocket, DTOInterface | None], Coroutine[Any, None, None]]:
    if resolved_data_dto:

        async def handle_receive(socket: WebSocket, dto: DTOInterface | None) -> Any:
            received_data = await socket.receive_data(mode=receive_mode)
            if isinstance(received_data, str):
                received_data = received_data.encode("utf-8")
            return cast("DTOInterface", dto).bytes_to_data_type(received_data)

    elif wants_receive_type is str:

        async def handle_receive(socket: WebSocket, dto: DTOInterface | None) -> Any:
            received_data = await socket.receive_data(mode=receive_mode)
            if isinstance(received_data, bytes):
                return received_data.decode("utf-8")
            return received_data

    elif wants_receive_type is bytes:

        async def handle_receive(socket: WebSocket, dto: DTOInterface | None) -> Any:
            received_data = await socket.receive_data(mode=receive_mode)
            if isinstance(received_data, str):
                return received_data.encode("utf-8")
            return received_data

    else:

        async def handle_receive(socket: WebSocket, dto: DTOInterface | None) -> Any:
            received_data = await socket.receive_data(mode=receive_mode)
            return decode_json(received_data)

    return handle_receive


def create_handle_send(
    resolved_return_dto: type[DTOInterface] | None,
    json_encoder: JsonEncoder,
    should_encode_to_json: bool,
    send_mode: WebSocketMode,
) -> Callable[[WebSocket, Any, DTOInterface | None], Coroutine[None, None, None]]:
    if resolved_return_dto:

        async def handle_send(socket: WebSocket, data_to_send: Any, dto: DTOInterface | None) -> None:
            data_to_send = json_encoder.encode(cast("DTOInterface", dto).data_to_encodable_type(data_to_send))
            await socket.send_data(data_to_send, send_mode)  # pyright: ignore

    elif should_encode_to_json:

        async def handle_send(socket: WebSocket, data_to_send: Any, dto: DTOInterface | None) -> None:
            data_to_send = json_encoder.encode(data_to_send)
            await socket.send_data(data_to_send, send_mode)  # pyright: ignore

    else:

        async def handle_send(socket: WebSocket, data_to_send: Any, dto: DTOInterface | None) -> None:
            await socket.send_data(data_to_send, send_mode)  # pyright: ignore

    return handle_send


def create_handler_function(
    listener_context: ListenerContext,
    on_accept: AsyncCallable | None,
    on_disconnect: AsyncCallable | None,
) -> Callable[..., Coroutine[None, None, None]]:
    async def handler_fn(socket: WebSocket, **kwargs: Any) -> None:
        await socket.accept()

        listener_callback = AsyncCallable(listener_context.listener_callback)
        ctx = ConnectionContext.from_connection(socket)
        data_dto = listener_context.resolved_data_dto(ctx) if listener_context.resolved_data_dto else None
        return_dto = listener_context.resolved_return_dto(ctx) if listener_context.resolved_return_dto else None

        if on_accept:
            await on_accept(socket)

        if listener_context.pass_socket:
            kwargs["socket"] = socket

        while True:
            try:
                received_data = await listener_context.handle_receive(socket, data_dto)
                data_to_send = await listener_callback(data=received_data, **kwargs)
                if listener_context.can_send_data:
                    await listener_context.handle_send(socket, data_to_send, return_dto)
            except WebSocketDisconnect:
                if on_disconnect:
                    await on_disconnect(socket)
                break

    return handler_fn


def create_handler_signature(callback_signature: inspect.Signature) -> inspect.Signature:
    """Creates a :class:`inspect.Signature` for the handler function for signature modelling.

    This is required for two reasons:

        1. the :class:`.handlers.WebsocketHandler` signature model cannot contain the ``data`` parameter, which is
            required for :class:`.handlers.websocket_listener` handlers.
        2. the :class;`.handlers.WebsocketHandler` signature model must include the ``socket`` parameter, which is
            optional for :class:`.handlers.websocket_listener` handlers.

    Args:
        callback_signature: The :class:`inspect.Signature` of the listener callback.

    Returns:
        The :class:`inspect.Signature` for the listener callback as required for signature modelling.
    """
    new_params = [p for p in callback_signature.parameters.values() if p.name not in {"data"}]
    if "socket" not in callback_signature.parameters:
        new_params.append(inspect.Parameter(name="socket", kind=inspect.Parameter.KEYWORD_ONLY, annotation="WebSocket"))
    return callback_signature.replace(parameters=new_params)



from __future__ import annotations

from functools import lru_cache
from http.cookies import _unquote as unquote_cookie
from typing import Any, Iterable
from urllib.parse import unquote

from fast_query_parsers import parse_query_string as fast_parse_query_string
from fast_query_parsers import parse_url_encoded_dict

__all__ = ("parse_cookie_string", "parse_headers", "parse_query_string", "parse_url_encoded_form_data")


@lru_cache(1024)
def parse_url_encoded_form_data(encoded_data: bytes) -> dict[str, Any]:
    """Parse an url encoded form data dict.

    Args:
        encoded_data: The encoded byte string.

    Returns:
        A parsed dict.
    """
    return parse_url_encoded_dict(encoded_data)


@lru_cache(1024)
def parse_query_string(query_string: bytes) -> tuple[tuple[str, Any], ...]:
    """Parse a query string into a tuple of key value pairs.

    Args:
        query_string: A query string.

    Returns:
        A tuple of key value pairs.
    """
    return tuple(fast_parse_query_string(query_string, "&"))


@lru_cache(1024)
def parse_cookie_string(cookie_string: str) -> dict[str, str]:
    """Parse a cookie string into a dictionary of values.

    Args:
        cookie_string: A cookie string.

    Returns:
        A string keyed dictionary of values
    """
    output: dict[str, str] = {}
    cookies = [cookie.split("=", 1) if "=" in cookie else ("", cookie) for cookie in cookie_string.split(";")]
    for k, v in filter(lambda x: x[0] or x[1], ((k.strip(), v.strip()) for k, v in cookies)):
        output[k] = unquote(unquote_cookie(v))
    return output


@lru_cache(1024)
def _parse_headers(headers: tuple[tuple[bytes, bytes], ...]) -> dict[str, str]:
    """Parse ASGI headers into a dict of string keys and values.

    Args:
        headers: A tuple of bytes two tuples.

    Returns:
        A string / string dict.
    """
    return {k.decode(): v.decode() for k, v in headers}


def parse_headers(headers: Iterable[tuple[bytes, bytes] | list[bytes]]) -> dict[str, str]:
    """Parse ASGI headers into a dict of string keys and values.

    Since the ASGI protocol only allows for lists (not tuples) which cannot be hashed,
    this function will convert the headers to a tuple of tuples before invoking the cached function.
    """
    return _parse_headers(tuple(tuple(h) for h in headers))



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any

from _decimal import Decimal
from polyfactory.exceptions import ParameterException
from polyfactory.field_meta import FieldMeta, Null

from litestar.openapi.spec import Example
from litestar.types import Empty
from litestar.utils import is_pydantic_model_instance

try:
    from polyfactory.factories.pydantic_factory import ModelFactory as Factory
except ImportError:
    from polyfactory.factories import DataclassFactory as Factory  # type: ignore[assignment]


if TYPE_CHECKING:
    from litestar._signature.field import SignatureField


def _normalize_example_value(value: Any) -> Any:
    """Normalize the example value to make it look a bit prettier."""
    if isinstance(value, (Decimal, float)):
        value = round(float(value), 2)
    if isinstance(value, Enum):
        value = value.value
    if is_pydantic_model_instance(value):
        value = value.dict()
    if isinstance(value, (list, set)):
        value = [_normalize_example_value(v) for v in value]
    if isinstance(value, dict):
        for k, v in value.items():
            value[k] = _normalize_example_value(v)
    return value


def _create_field_meta(field: SignatureField) -> FieldMeta:
    return FieldMeta(
        name=field.name,
        annotation=field.field_type,
        constraints={"constant": field.is_const},
        default=field.default_value if field.default_value is not Empty else Null,
        children=[_create_field_meta(child) for child in field.children] if field.children else None,
    )


def create_examples_for_field(field: SignatureField) -> list[Example]:
    """Create an OpenAPI Example instance.

    Args:
        field: A signature field.

    Returns:
        A list including a single example.
    """
    try:
        field_meta = _create_field_meta(field)
        value = _normalize_example_value(Factory.get_field_value(field_meta))
        return [Example(description=f"Example {field.name} value", value=value)]
    except ParameterException:  # pragma: no cover
        return []



from __future__ import annotations

from contextlib import contextmanager
from http.cookiejar import CookieJar
from typing import TYPE_CHECKING, Any, Generator, Generic, Mapping, TypeVar, cast
from warnings import warn

from anyio.from_thread import BlockingPortal, start_blocking_portal

from litestar.connection import ASGIConnection
from litestar.datastructures import MutableScopeHeaders
from litestar.exceptions import (
    ImproperlyConfiguredException,
)
from litestar.types import AnyIOBackend, ASGIApp, HTTPResponseStartEvent

if TYPE_CHECKING:
    from httpx._types import CookieTypes

    from litestar.middleware.session.base import BaseBackendConfig, BaseSessionBackend
    from litestar.middleware.session.client_side import ClientSideSessionBackend
from httpx import Cookies, Request, Response

T = TypeVar("T", bound=ASGIApp)


def fake_http_send_message(headers: MutableScopeHeaders) -> HTTPResponseStartEvent:
    headers.setdefault("content-type", "application/text")
    return HTTPResponseStartEvent(type="http.response.start", status=200, headers=headers.headers)


def fake_asgi_connection(app: ASGIApp, cookies: dict[str, str]) -> ASGIConnection[Any, Any, Any, Any]:
    scope = {
        "type": "http",
        "path": "/",
        "raw_path": b"/",
        "root_path": "",
        "scheme": "http",
        "query_string": b"",
        "client": ("testclient", 50000),
        "server": ("testserver", 80),
        "method": "GET",
        "http_version": "1.1",
        "extensions": {"http.response.template": {}},
        "app": app,
        "state": {},
        "path_params": {},
        "route_handler": None,
        "_cookies": cookies,
    }
    return ASGIConnection[Any, Any, Any, Any](
        scope=scope,  # type: ignore[arg-type]
    )


class BaseTestClient(Generic[T]):
    __test__ = False
    blocking_portal: BlockingPortal

    __slots__ = (
        "app",
        "base_url",
        "backend",
        "backend_options",
        "session_config",
        "_session_backend",
        "cookies",
    )

    def __init__(
        self,
        app: T,
        base_url: str = "http://testserver.local",
        backend: AnyIOBackend = "asyncio",
        backend_options: Mapping[str, Any] | None = None,
        session_config: BaseBackendConfig | None = None,
        cookies: CookieTypes | None = None,
    ):
        if "." not in base_url:
            warn(
                f"The base_url {base_url!r} might cause issues. Try adding a domain name such as .local: "
                f"'{base_url}.local'",
                UserWarning,
                stacklevel=1,
            )
        self._session_backend: BaseSessionBackend | None = None
        if session_config:
            self._session_backend = session_config._backend_class(config=session_config)
        self.app = app
        self.base_url = base_url
        self.backend = backend
        self.backend_options = backend_options
        self.cookies = cookies

    @property
    def session_backend(self) -> BaseSessionBackend:
        if not self._session_backend:
            raise ImproperlyConfiguredException(
                "Session has not been initialized for this TestClient instance. You can"
                "do so by passing a configuration object to TestClient: TestClient(app=app, session_config=...)"
            )
        return self._session_backend

    @contextmanager
    def portal(self) -> Generator[BlockingPortal, None, None]:
        """Get a BlockingPortal.

        Returns:
            A contextmanager for a BlockingPortal.
        """
        if hasattr(self, "blocking_portal"):
            yield self.blocking_portal
        else:
            with start_blocking_portal(
                backend=self.backend, backend_options=dict(self.backend_options or {})
            ) as portal:
                yield portal

    @staticmethod
    def _create_session_cookies(backend: ClientSideSessionBackend, data: dict[str, Any]) -> dict[str, str]:
        encoded_data = backend.dump_data(data=data)
        return {cookie.key: cast("str", cookie.value) for cookie in backend._create_session_cookies(encoded_data)}

    async def _set_session_data(self, data: dict[str, Any]) -> None:
        mutable_headers = MutableScopeHeaders()
        await self.session_backend.store_in_message(
            scope_session=data,
            message=fake_http_send_message(mutable_headers),
            connection=fake_asgi_connection(
                app=self.app,
                cookies=dict(self.cookies),  # type: ignore[arg-type]
            ),
        )
        response = Response(200, request=Request("GET", self.base_url), headers=mutable_headers.headers)

        cookies = Cookies(CookieJar())
        cookies.extract_cookies(response)
        self.cookies.update(cookies)  # type: ignore[union-attr]

    async def _get_session_data(self) -> dict[str, Any]:
        return await self.session_backend.load_from_connection(
            connection=fake_asgi_connection(
                app=self.app,
                cookies=dict(self.cookies),  # type: ignore[arg-type]
            ),
        )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Callable, Generic, TypeVar, cast

from litestar.constants import HTTP_DISCONNECT, HTTP_RESPONSE_START, WEBSOCKET_CLOSE, WEBSOCKET_DISCONNECT
from litestar.exceptions import ImproperlyConfiguredException
from litestar.types import Empty
from litestar.utils import get_litestar_scope_state, set_litestar_scope_state
from litestar.utils.dataclass import simple_asdict

from .engine import EngineConfig

if TYPE_CHECKING:
    from typing import Any

    from sqlalchemy import Connection, Engine
    from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, AsyncSession, async_sessionmaker
    from sqlalchemy.orm import Mapper, Query, Session, sessionmaker
    from sqlalchemy.orm.session import JoinTransactionMode
    from sqlalchemy.sql import TableClause

    from litestar.datastructures.state import State
    from litestar.types import BeforeMessageSendHookHandler, EmptyType, Scope

__all__ = (
    "SESSION_SCOPE_KEY",
    "SESSION_TERMINUS_ASGI_EVENTS",
    "GenericSQLAlchemyConfig",
    "GenericSessionConfig",
)

SESSION_SCOPE_KEY = "_sqlalchemy_db_session"
SESSION_TERMINUS_ASGI_EVENTS = {HTTP_RESPONSE_START, HTTP_DISCONNECT, WEBSOCKET_DISCONNECT, WEBSOCKET_CLOSE}

ConnectionT = TypeVar("ConnectionT", bound="Connection | AsyncConnection")
EngineT = TypeVar("EngineT", bound="Engine | AsyncEngine")
SessionT = TypeVar("SessionT", bound="Session | AsyncSession")
SessionMakerT = TypeVar("SessionMakerT", bound="sessionmaker | async_sessionmaker")


@dataclass
class GenericSessionConfig(Generic[ConnectionT, EngineT, SessionT]):
    """SQLAlchemy async session config."""

    autobegin: bool | EmptyType = Empty
    """Automatically start transactions when database access is requested by an operation."""
    autoflush: bool | EmptyType = Empty
    """When ``True``, all query operations will issue a flush call to this :class:`Session <sqlalchemy.orm.Session>`
    before proceeding"""
    bind: EngineT | ConnectionT | None | EmptyType = Empty
    """The :class:`Engine <sqlalchemy.engine.Engine>` or :class:`Connection <sqlalchemy.engine.Connection>` that new
    :class:`Session <sqlalchemy.orm.Session>` objects will be bound to."""
    binds: dict[type[Any] | Mapper[Any] | TableClause | str, EngineT | ConnectionT] | None | EmptyType = Empty
    """A dictionary which may specify any number of :class:`Engine <sqlalchemy.engine.Engine>` or :class:`Connection
    <sqlalchemy.engine.Connection>` objects as the source of connectivity for SQL operations on a per-entity basis. The
    keys of the dictionary consist of any series of mapped classes, arbitrary Python classes that are bases for mapped
    classes, :class:`Table <sqlalchemy.schema.Table>` objects and :class:`Mapper <sqlalchemy.orm.Mapper>` objects. The
    values of the dictionary are then instances of :class:`Engine <sqlalchemy.engine.Engine>` or less commonly
    :class:`Connection <sqlalchemy.engine.Connection>` objects."""
    class_: type[SessionT] | EmptyType = Empty
    """Class to use in order to create new :class:`Session <sqlalchemy.orm.Session>` objects."""
    expire_on_commit: bool | EmptyType = Empty
    """If ``True``, all instances will be expired after each commit."""
    info: dict[str, Any] | None | EmptyType = Empty
    """Optional dictionary of information that will be available via the
    :attr:`Session.info <sqlalchemy.orm.Session.info>`"""
    join_transaction_mode: JoinTransactionMode | EmptyType = Empty
    """Describes the transactional behavior to take when a given bind is a Connection that has already begun a
    transaction outside the scope of this Session; in other words the
    :attr:`Connection.in_transaction() <sqlalchemy.Connection.in_transaction>` method returns True."""
    query_cls: type[Query] | None | EmptyType = Empty
    """Class which should be used to create new Query objects, as returned by the
    :attr:`Session.query() <sqlalchemy.orm.Session.query>` method."""
    twophase: bool | EmptyType = Empty
    """When ``True``, all transactions will be started as a “two phase” transaction, i.e. using the “two phase”
    semantics of the database in use along with an XID. During a :attr:`commit() <sqlalchemy.orm.Session.commit>`, after
    :attr:`flush() <sqlalchemy.orm.Session.flush>` has been issued for all attached databases, the
    :attr:`TwoPhaseTransaction.prepare() <sqlalchemy.engine.TwoPhaseTransaction.prepare>` method on each database`s
    :class:`TwoPhaseTransaction <sqlalchemy.engine.TwoPhaseTransaction>` will be called. This allows each database to
    roll back the entire transaction, before each transaction is committed."""


@dataclass
class GenericSQLAlchemyConfig(Generic[EngineT, SessionT, SessionMakerT]):
    """Common SQLAlchemy Configuration."""

    create_engine_callable: Callable[[str], EngineT]
    """Callable that creates an :class:`AsyncEngine <sqlalchemy.ext.asyncio.AsyncEngine>` instance or instance of its
    subclass.
    """
    session_config: GenericSessionConfig
    """Configuration options for either the :class:`async_sessionmaker <sqlalchemy.ext.asyncio.async_sessionmaker>`
    or :class:`sessionmaker <sqlalchemy.orm.sessionmaker>`.
    """
    session_maker_class: type[sessionmaker] | type[async_sessionmaker]
    """Sessionmaker class to use."""
    before_send_handler: BeforeMessageSendHookHandler
    """Handler to call before the ASGI message is sent.

    The handler should handle closing the session stored in the ASGI scope, if its still open, and committing and
    uncommitted data.
    """
    connection_string: str | None = field(default=None)
    """Database connection string in one of the formats supported by SQLAlchemy.

    Notes:
        - For async connections, the connection string must include the correct async prefix.
          e.g. ``'postgresql+asyncpg://...'`` instead of ``'postgresql://'``, and for sync connections its the opposite.

    """
    engine_dependency_key: str = "db_engine"
    """Key to use for the dependency injection of database engines."""
    session_dependency_key: str = "db_session"
    """Key to use for the dependency injection of database sessions."""
    engine_app_state_key: str = "db_engine"
    """Key under which to store the SQLAlchemy engine in the application :class:`State <.datastructures.State>`
    instance.
    """
    engine_config: EngineConfig = field(default_factory=EngineConfig)
    """Configuration for the SQLAlchemy engine.

    The configuration options are documented in the SQLAlchemy documentation.
    """
    session_maker_app_state_key: str = "session_maker_class"
    """Key under which to store the SQLAlchemy :class:`sessionmaker <sqlalchemy.orm.sessionmaker>` in the application
    :class:`State <.datastructures.State>` instance.
    """
    session_maker: Callable[[], SessionT] | None = None
    """Callable that returns a session.

    If provided, the plugin will use this rather than instantiate a sessionmaker.
    """
    engine_instance: EngineT | None = None
    """Optional engine to use.

    If set, the plugin will use the provided instance rather than instantiate an engine.
    """

    def __post_init__(self) -> None:
        if self.connection_string is not None and self.engine_instance is not None:
            raise ImproperlyConfiguredException("Only one of 'connection_string' or 'engine_instance' can be provided.")

    @property
    def engine_config_dict(self) -> dict[str, Any]:
        """Return the engine configuration as a dict.

        Returns:
            A string keyed dict of config kwargs for the SQLAlchemy :func:`create_engine <sqlalchemy.create_engine>`
            function.
        """
        return simple_asdict(self.engine_config, exclude_empty=True)

    @property
    def session_config_dict(self) -> dict[str, Any]:
        """Return the session configuration as a dict.

        Returns:
            A string keyed dict of config kwargs for the SQLAlchemy :class:`sessionmaker <sqlalchemy.orm.sessionmaker>`
            class.
        """
        return simple_asdict(self.session_config, exclude_empty=True)

    @property
    def signature_namespace(self) -> dict[str, Any]:
        """Return the plugin's signature namespace.

        Returns:
            A string keyed dict of names to be added to the namespace for signature forward reference resolution.
        """
        return {}  # pragma: no cover

    def create_engine(self) -> EngineT:
        """Return an engine. If none exists yet, create one.

        Returns:
            Getter that returns the engine instance used by the plugin.
        """
        if self.engine_instance:
            return self.engine_instance

        if self.connection_string is None:
            raise ImproperlyConfiguredException("One of 'connection_string' or 'engine_instance' must be provided.")

        engine_config = self.engine_config_dict
        try:
            return self.create_engine_callable(self.connection_string, **engine_config)
        except ValueError:
            # likely due to a dialect that doesn't support json type
            del engine_config["json_deserializer"]
            del engine_config["json_serializer"]
            return self.create_engine_callable(self.connection_string, **engine_config)

    def create_session_maker(self) -> Callable[[], SessionT]:
        """Get a session maker. If none exists yet, create one.

        Returns:
            Session factory used by the plugin.
        """
        if self.session_maker:
            return self.session_maker

        session_kws = self.session_config_dict
        if session_kws.get("bind") is None:
            session_kws["bind"] = self.create_engine()
        return self.session_maker_class(**session_kws)

    def provide_engine(self, state: State) -> EngineT:
        """Create an engine instance.

        Args:
            state: The ``Litestar.state`` instance.

        Returns:
            An engine instance.
        """
        return cast("EngineT", state.get(self.engine_app_state_key))

    def provide_session(self, state: State, scope: Scope) -> SessionT:
        """Create a session instance.

        Args:
            state: The ``Litestar.state`` instance.
            scope: The current connection's scope.

        Returns:
            A session instance.
        """
        session = cast("SessionT | None", get_litestar_scope_state(scope, SESSION_SCOPE_KEY))
        if session is None:
            session_maker = cast("Callable[[], SessionT]", state[self.session_maker_app_state_key])
            session = session_maker()
            set_litestar_scope_state(scope, SESSION_SCOPE_KEY, session)
        return session

    def create_app_state_items(self) -> dict[str, Any]:
        """Key/value pairs to be stored in application state."""
        return {
            self.engine_app_state_key: self.create_engine(),
            self.session_maker_app_state_key: self.create_session_maker(),
        }

    def update_app_state(self, state: State) -> None:
        """Set the app state with engine and session.

        Args:
            state: The ``Litestar.state`` instance.
        """
        state.update(self.create_app_state_items())



from __future__ import annotations

from datetime import datetime
from re import Pattern
from typing import TYPE_CHECKING, Any

from litestar._openapi.schema_generation.utils import sort_schemas_and_references
from litestar.exceptions import MissingDependencyException
from litestar.openapi.spec.enums import OpenAPIFormat, OpenAPIType
from litestar.openapi.spec.schema import Schema

if TYPE_CHECKING:
    from litestar.plugins import OpenAPISchemaPluginProtocol

if TYPE_CHECKING:
    from litestar._signature.field import SignatureField

    try:
        from pydantic import (
            ConstrainedBytes,
            ConstrainedDate,
            ConstrainedDecimal,
            ConstrainedFloat,
            ConstrainedFrozenSet,
            ConstrainedInt,
            ConstrainedList,
            ConstrainedSet,
            ConstrainedStr,
        )
    except ImportError:
        ConstrainedBytes = Any  # type: ignore
        ConstrainedDate = Any  # type: ignore
        ConstrainedDecimal = Any  # type: ignore
        ConstrainedFloat = Any  # type: ignore
        ConstrainedFrozenSet = Any  # type: ignore
        ConstrainedInt = Any  # type: ignore
        ConstrainedList = Any  # type: ignore
        ConstrainedSet = Any  # type: ignore
        ConstrainedStr = Any  # type: ignore


__all__ = (
    "create_collection_constrained_field_schema",
    "create_constrained_field_schema",
    "create_date_constrained_field_schema",
    "create_numerical_constrained_field_schema",
    "create_string_constrained_field_schema",
)


def create_numerical_constrained_field_schema(
    field_type: type[ConstrainedFloat] | type[ConstrainedInt] | type[ConstrainedDecimal],
) -> Schema:
    """Create Schema from Constrained Int/Float/Decimal field."""
    schema = Schema(type=OpenAPIType.INTEGER if issubclass(field_type, int) else OpenAPIType.NUMBER)
    if field_type.le is not None:
        schema.maximum = float(field_type.le)
    if field_type.lt is not None:
        schema.exclusive_maximum = float(field_type.lt)
    if field_type.ge is not None:
        schema.minimum = float(field_type.ge)
    if field_type.gt is not None:
        schema.exclusive_minimum = float(field_type.gt)
    if field_type.multiple_of is not None:
        schema.multiple_of = float(field_type.multiple_of)
    return schema


def create_date_constrained_field_schema(field_type: type[ConstrainedDate]) -> Schema:
    """Create Schema from Constrained Date Field."""
    schema = Schema(type=OpenAPIType.STRING, format=OpenAPIFormat.DATE)
    if field_type.le is not None:
        schema.maximum = float(datetime.combine(field_type.le, datetime.min.time()).timestamp())
    if field_type.lt is not None:
        schema.exclusive_maximum = float(datetime.combine(field_type.lt, datetime.min.time()).timestamp())
    if field_type.ge is not None:
        schema.minimum = float(datetime.combine(field_type.ge, datetime.min.time()).timestamp())
    if field_type.gt is not None:
        schema.exclusive_minimum = float(datetime.combine(field_type.gt, datetime.min.time()).timestamp())
    return schema


def create_string_constrained_field_schema(field_type: type[ConstrainedStr] | type[ConstrainedBytes]) -> Schema:
    """Create Schema from Constrained Str/Bytes field."""
    schema = Schema(type=OpenAPIType.STRING)
    if field_type.min_length:
        schema.min_length = field_type.min_length
    if field_type.max_length:
        schema.max_length = field_type.max_length
    if hasattr(field_type, "regex") and isinstance(field_type.regex, Pattern):
        schema.pattern = field_type.regex.pattern
    if field_type.to_lower:
        schema.description = "must be in lower case"
    return schema


def create_collection_constrained_field_schema(
    field_type: type[ConstrainedList] | type[ConstrainedSet] | type[ConstrainedFrozenSet],
    children: tuple[SignatureField, ...] | None,
    plugins: list[OpenAPISchemaPluginProtocol],
    schemas: dict[str, Schema],
) -> Schema:
    """Create Schema from Constrained List/Set field.

    Args:
        field_type: A constrained field type.
        children: Any child fields.
        plugins: A list of plugins.
        schemas: A mapping of namespaces to schemas - this mapping is used in the OA components section.

    Returns:
        A schema instance.
    """
    import pydantic

    from litestar._openapi.schema_generation import create_schema

    schema = Schema(type=OpenAPIType.ARRAY)
    if field_type.min_items:
        schema.min_items = field_type.min_items
    if field_type.max_items:
        schema.max_items = field_type.max_items
    if issubclass(field_type, (pydantic.ConstrainedSet, pydantic.ConstrainedFrozenSet)):
        schema.unique_items = True
    if children:
        items = [
            create_schema(field=sub_field, generate_examples=False, plugins=plugins, schemas=schemas)
            for sub_field in children
        ]
        if len(items) > 1:
            schema.items = Schema(one_of=sort_schemas_and_references(items))
        else:
            schema.items = items[0]
    else:
        from litestar._signature.field import SignatureField

        schema.items = create_schema(
            field=SignatureField.create(field_type=field_type.item_type, name=f"{field_type.__name__}Field"),
            generate_examples=False,
            plugins=plugins,
            schemas=schemas,
        )
    return schema


def create_constrained_field_schema(
    field_type: type[ConstrainedBytes]
    | type[ConstrainedDate]
    | type[ConstrainedDecimal]
    | type[ConstrainedFloat]
    | type[ConstrainedFrozenSet]
    | type[ConstrainedInt]
    | type[ConstrainedList]
    | type[ConstrainedSet]
    | type[ConstrainedStr],
    children: tuple[SignatureField, ...] | None,
    plugins: list[OpenAPISchemaPluginProtocol],
    schemas: dict[str, Schema],
) -> Schema:
    """Create Schema for Pydantic Constrained fields (created using constr(), conint() and so forth, or by subclassing
    Constrained*)

    Args:
        field_type: A constrained field type.
        children: Any children.
        plugins: A list of plugins.
        schemas: A mapping of namespaces to schemas - this mapping is used in the OA components section.

    Returns:
        A schema instance.

    """

    try:
        import pydantic
    except ImportError as e:
        raise MissingDependencyException("pydantic") from e

    if issubclass(field_type, (pydantic.ConstrainedFloat, pydantic.ConstrainedInt, pydantic.ConstrainedDecimal)):
        return create_numerical_constrained_field_schema(field_type=field_type)
    if issubclass(field_type, (pydantic.ConstrainedStr, pydantic.ConstrainedBytes)):
        return create_string_constrained_field_schema(field_type=field_type)
    if issubclass(field_type, pydantic.ConstrainedDate):
        return create_date_constrained_field_schema(field_type=field_type)
    return create_collection_constrained_field_schema(
        field_type=field_type, children=tuple(children) if children else None, plugins=plugins, schemas=schemas
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Literal,
    TypeVar,
)

from litestar.constants import DEFAULT_CHUNK_SIZE
from litestar.exceptions import ImproperlyConfiguredException
from litestar.file_system import BaseLocalFileSystem
from litestar.response import FileResponse, RedirectResponse, StreamingResponse, TemplateResponse

__all__ = ("File", "Redirect", "ResponseContainer", "Stream", "Template")


if TYPE_CHECKING:
    import os

    from litestar.app import Litestar
    from litestar.background_tasks import BackgroundTask, BackgroundTasks
    from litestar.connection import Request
    from litestar.datastructures.cookie import Cookie
    from litestar.datastructures.headers import ETag
    from litestar.enums import MediaType
    from litestar.types import FileInfo
    from litestar.types.composite_types import PathType, StreamType

R_co = TypeVar("R_co", covariant=True)


class ResponseContainer(ABC, Generic[R_co]):
    """Generic response container."""

    background: BackgroundTask | BackgroundTasks | None
    """A :class:`BackgroundTask <.background_tasks.BackgroundTask>` instance or.

    :class:`BackgroundTasks <.background_tasks.BackgroundTasks>` to execute after the response is finished.
    Defaults to None.
    """
    headers: dict[str, Any]
    """A string/string dictionary of response headers.

    Header keys are insensitive. Defaults to None.
    """
    cookies: list[Cookie]
    """A list of Cookie instances to be set under the response ``Set-Cookie`` header.

    Defaults to None.
    """
    media_type: MediaType | str | None
    """If defined, overrides the media type configured in the route decorator."""
    encoding: str
    """The encoding to be used for the response headers."""

    @abstractmethod
    def to_response(
        self,
        headers: dict[str, Any],
        media_type: MediaType | str,
        status_code: int,
        app: Litestar,
        request: Request,
    ) -> R_co:  # pragma: no cover
        """Abstract method that should be implemented by subclasses.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the :class:`MediaType <.enums.MediaType>` enum.
            status_code: A response status code.
            app: The :class:`Litestar <.app.Litestar>` application instance.
            request: A :class:`Request <.connection.Request>` instance.

        Returns:
            A Response Object
        """
        raise NotImplementedError("not implemented")


@dataclass
class File(ResponseContainer[FileResponse]):
    """Container type for returning File responses."""

    path: PathType
    """Path to the file to send."""
    background: BackgroundTask | BackgroundTasks | None = field(default=None)
    """A :class:`BackgroundTask <.background_tasks.BackgroundTask>` instance or.

    :class:`BackgroundTasks <.background_tasks.BackgroundTasks>` to execute after the response is finished.
    Defaults to None.
    """
    headers: dict[str, Any] = field(default_factory=dict)
    """A string/string dictionary of response headers.

    Header keys are insensitive. Defaults to None.
    """
    cookies: list[Cookie] = field(default_factory=list)
    """A list of Cookie instances to be set under the response ``Set-Cookie`` header.

    Defaults to None.
    """
    media_type: MediaType | str | None = field(default=None)
    """If defined, overrides the media type configured in the route decorator."""
    encoding: str = field(default="utf-8")
    """The encoding to be used for the response headers."""
    filename: str | None = field(default=None)
    """An optional filename to set in the header."""
    stat_result: os.stat_result | None = field(default=None)
    """An optional result of calling 'os.stat'.

    If not provided, this will be done by the response constructor.
    """
    chunk_size: int = field(default=DEFAULT_CHUNK_SIZE)
    """The size of chunks to use when streaming the file."""
    content_disposition_type: Literal["attachment", "inline"] = field(default="attachment")
    """The type of the 'Content-Disposition'.

    Either ``inline`` or ``attachment``.
    """
    etag: ETag | None = field(default=None)
    """An optional :class:`ETag <.datastructures.ETag>` instance.

    If not provided, an etag will be automatically generated.
    """
    file_system: Any = field(default_factory=BaseLocalFileSystem)
    """The file_system spec to use loading the file.

    Notes:
        - A file_system is a class that adheres to the :class:`FileSystemProtocol <.types.FileSystemProtocol>`.
        - You can use any of the file systems exported from the
          `fsspec <https://filesystem-spec.readthedocs.io/en/latest/>`_ library for this purpose.
    """
    file_info: FileInfo | None = field(default=None)
    """The output of calling :meth:`file_system.info <.types.FileSystemProtocol.info>`, equivalent to providing a
    :class`os.stat_result`."""

    def __post_init__(self) -> None:
        if not (
            callable(getattr(self.file_system, "info", None)) and callable(getattr(self.file_system, "open", None))
        ):
            raise ImproperlyConfiguredException("file_system must adhere to the FileSystemProtocol type")

        if not self.stat_result:
            self.stat_result = Path(self.path).stat()

    def to_response(
        self,
        headers: dict[str, Any],
        media_type: MediaType | str | None,
        status_code: int,
        app: Litestar,
        request: Request,
    ) -> FileResponse:
        """Create a FileResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the :class:`MediaType <.enums.MediaType>` enum.
            status_code: A response status code.
            app: The :class:`Litestar <.app.Litestar>` application instance.
            request: A :class:`Request <.connection.Request>` instance.

        Returns:
            A :class:`FileResponse <.response.FileResponse>` instance
        """
        return FileResponse(
            background=self.background,
            chunk_size=self.chunk_size,
            content_disposition_type=self.content_disposition_type,
            encoding=self.encoding,
            etag=self.etag,
            file_info=self.file_info,
            file_system=self.file_system,
            filename=self.filename,
            headers=headers,
            media_type=media_type,
            path=self.path,
            stat_result=self.stat_result,
            status_code=status_code,
        )


@dataclass
class Redirect(ResponseContainer[RedirectResponse]):
    """Container type for returning Redirect responses."""

    path: str
    """Redirection path."""
    background: BackgroundTask | BackgroundTasks | None = field(default=None)
    """A :class:`BackgroundTask <.background_tasks.BackgroundTask>` instance or
    :class:`BackgroundTasks <.background_tasks.BackgroundTasks>` to execute after the response is finished.
    Defaults to None.
    """
    headers: dict[str, Any] = field(default_factory=dict)
    """A string/string dictionary of response headers.

    Header keys are insensitive. Defaults to None.
    """
    cookies: list[Cookie] = field(default_factory=list)
    """A list of Cookie instances to be set under the response ``Set-Cookie`` header.

    Defaults to None.
    """
    media_type: MediaType | str | None = field(default=None)
    """If defined, overrides the media type configured in the route decorator."""
    encoding: str = field(default="utf-8")
    """The encoding to be used for the response headers."""
    status_code: Literal[301, 302, 303, 307, 308] | None = None
    """Redirect status code"""

    def to_response(  # type: ignore[override]
        self,
        headers: dict[str, Any],
        media_type: MediaType | str,
        status_code: Literal[301, 302, 303, 307, 308],
        app: Litestar,
        request: Request,
    ) -> RedirectResponse:
        """Create a RedirectResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the :class:`MediaType <.enums.MediaType>` enum.
            status_code: A response status code.
            app: The :class:`Litestar <.app.Litestar>` application instance.
            request: A :class:`Request <.connection.Request>` instance.

        Returns:
            A :class:`RedirectResponse <.response.RedirectResponse>` instance
        """
        return RedirectResponse(
            background=self.background,
            encoding=self.encoding,
            headers=headers,
            status_code=self.status_code or status_code,
            url=self.path,
        )


@dataclass
class Stream(ResponseContainer[StreamingResponse]):
    """Container type for returning Stream responses."""

    iterator: StreamType[str | bytes] | Callable[[], StreamType[str | bytes]]
    """Iterator, Iterable,Generator or async Iterator, Iterable or Generator returning chunks to stream."""
    background: BackgroundTask | BackgroundTasks | None = field(default=None)
    """A :class:`BackgroundTask <.background_tasks.BackgroundTask>` instance or.

    :class:`BackgroundTasks <.background_tasks.BackgroundTasks>` to execute after the response is finished.
    Defaults to None.
    """
    headers: dict[str, Any] = field(default_factory=dict)
    """A string/string dictionary of response headers.

    Header keys are insensitive. Defaults to None.
    """
    cookies: list[Cookie] = field(default_factory=list)
    """A list of Cookie instances to be set under the response ``Set-Cookie`` header.

    Defaults to None.
    """
    media_type: MediaType | str | None = field(default=None)
    """If defined, overrides the media type configured in the route decorator."""
    encoding: str = field(default="utf-8")
    """The encoding to be used for the response headers."""

    def __post_init__(self) -> None:
        """Set the iterator value by ensuring that the return value is iterable.

        Returns:
            A sync or async iterable.
        """

        if not isinstance(self.iterator, (Iterable, Iterator, AsyncIterable, AsyncIterator)) and callable(
            self.iterator
        ):
            self.iterator = self.iterator()

        if not isinstance(self.iterator, (Iterable, Iterator, AsyncIterable, AsyncIterator)):
            raise ImproperlyConfiguredException(
                "iterator must be either an iterator or a callable that returns an iterator"
            )

    def to_response(
        self,
        headers: dict[str, Any],
        media_type: MediaType | str,
        status_code: int,
        app: Litestar,
        request: Request,
    ) -> StreamingResponse:
        """Create a StreamingResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the :class:`MediaType <.enums.MediaType>` enum.
            status_code: A response status code.
            app: The :class:`Litestar <.app.Litestar>` application instance.
            request: A :class:`Request <.connection.Request>` instance.

        Returns:
            A :class:`StreamingResponse <.response.StreamingResponse>` instance
        """

        return StreamingResponse(
            background=self.background,
            content=self.iterator if isinstance(self.iterator, (Iterable, AsyncIterable)) else self.iterator(),
            encoding=self.encoding,
            headers=headers,
            media_type=media_type,
            status_code=status_code,
        )


@dataclass
class Template(ResponseContainer[TemplateResponse]):
    """Container type for returning Template responses."""

    name: str
    """Path-like name for the template to be rendered, e.g. "index.html"."""
    context: dict[str, Any] = field(default_factory=dict)
    """A dictionary of key/value pairs to be passed to the temple engine's render method.

    Defaults to None.
    """
    background: BackgroundTask | BackgroundTasks | None = field(default=None)
    """A :class:`BackgroundTask <.background_tasks.BackgroundTask>` instance or
    :class:`BackgroundTasks <.background_tasks.BackgroundTasks>` to execute after the response is finished. Defaults to
    None.
    """
    headers: dict[str, Any] = field(default_factory=dict)
    """A string/string dictionary of response headers. Header keys are insensitive. Defaults to None."""
    cookies: list[Cookie] = field(default_factory=list)
    """A list of Cookie instances to be set under the response ``Set-Cookie`` header. Defaults to None. """
    media_type: MediaType | str | None = field(default=None)
    """If defined, overrides the media type configured in the route decorator."""
    encoding: str = field(default="utf-8")
    """The encoding to be used for the response headers."""

    def to_response(
        self,
        headers: dict[str, Any],
        media_type: MediaType | str,
        status_code: int,
        app: Litestar,
        request: Request,
    ) -> TemplateResponse:
        """Create a TemplateResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the :class:`MediaType <.enums.MediaType>` enum.
            status_code: A response status code.
            app: The :class:`Litestar <.app.Litestar>` application instance.
            request: A :class:`Request <.connection.Request>` instance.

        Raises:
            ImproperlyConfiguredException: If ``Litestar.template_engine`` is not configured.

        Returns:
            A :class:`TemplateResponse <.response.TemplateResponse>` instance
        """
        if not app.template_engine:
            raise ImproperlyConfiguredException("Template engine is not configured")

        return TemplateResponse(
            background=self.background,
            context=self.create_template_context(request=request),
            encoding=self.encoding,
            headers=headers,
            status_code=status_code,
            template_engine=app.template_engine,
            template_name=self.name,
            media_type=media_type,
        )

    def create_template_context(self, request: Request) -> dict[str, Any]:
        """Create a context object for the template.

        Args:
            request: A :class:`Request <.connection.Request>` instance.

        Returns:
            A dictionary holding the template context
        """
        csrf_token = request.scope.get("_csrf_token", "")
        return {
            **self.context,
            "request": request,
            "csrf_input": f'<input type="hidden" name="_csrf_token" value="{csrf_token}" />',
        }
